from setuptools import setup

setup(
    name='NlpToolkit-Hmm',
    version='1.0.4',
    packages=['Hmm'],
    url='https://github.com/olcaytaner/Hmm-Py',
    license='',
    author='olcaytaner',
    author_email='olcaytaner@isikun.edu.tr',
    description='Hidden Markov Model Library',
    install_requires=['NlpToolkit-Math', 'NlpToolkit-DataStructure']
)
